/* vim:ts=4 
 */

#include "misc.h"
#include "bmf.h"
#include <SDL/SDL.h>
#include <SDL/SDL_gfxPrimitives.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#define READ(buf, count)\
	if ((count) != read(f, buf, (count))) {\
		debug(sev_error, "bmf", "read: %s", strerror(errno));\
		goto err;\
	}

struct bmf_font *
bmf_load(const char *fname)
{
	int f;
	unsigned char buf[0x10000];
	struct bmf_font *font = NULL;
	/* in bmf, color 0 is meant to be transparent. i start
	 * reading palette 3 bytes after and put pink (rgb = 0xff00ff)
	 * in the first color. */
	char palette[771];
	int rgb_entries;
	char c;
	int i;
	off_t charsoffset;
	int totalwidth, maxheight;
	struct bmf_char *ch;
	SDL_Surface *tmp_surface;
	
	if (-1 == (f = open(fname, O_RDONLY))) {
		debug(sev_error, "bmf", "open: %s", strerror(errno));
		goto err;
	}

		/* Read header */
	READ(buf, 17);
	if (buf[0] != 0xE1 || buf[1] != 0xE6 || buf[2] != 0xD5 || buf[3] != 0x1A) {
		debug(sev_error, "bmf", "Bad magic");
		goto err;
	}
	font = xmalloc(sizeof(struct bmf_font));
	font->chars = NULL;
	font->info = NULL;
	font->surface = NULL;
	font->version = buf[4];
	font->line_height = buf[5];
	font->size_over = (signed)buf[6];
	font->size_under = (signed)buf[7];
	font->add_space = (signed)buf[8];
	font->size_inner = (signed)buf[9];
	font->colors = buf[10];
	font->highest_color = buf[11];
	rgb_entries = buf[16];
	READ(palette + 3, 3 * rgb_entries);
	for (i = 0; i < sizeof(palette); i++)
		palette[i] = palette[i] * 4;
	palette[0] = 0xff; palette[1] = 0x0; palette[2] = 0xff;
	READ(&c, 1);
	font->info = xmalloc(c + 1);
	READ(font->info, c);
	font->info[(int)c] = '\0';
	READ(buf, 2);
	font->numchars = buf[0] | (buf[1] << 8);
	font->chars = xmalloc(font->numchars * sizeof(struct bmf_char));
	for (i = 0; i < 256; i++)
		font->charmap[i] = -1;

		/* Read char definitions and get sum of widths */
	charsoffset = lseek(f, 0, SEEK_CUR);
	totalwidth = maxheight = 0;
	ch = font->chars;
	for (i = 0; i < font->numchars; i++) {
		READ(buf, 6);
		font->charmap[(int)buf[0]] = i;
		ch->w = buf[1];
		ch->h = buf[2];
		if (ch->h > maxheight)
			maxheight = ch->h;
		ch->relx = (signed)buf[3];
		ch->rely = (signed)buf[4];
		ch->shift = buf[5];
		ch->offset = totalwidth;
		totalwidth += ch->w;
		lseek(f, ch->w * ch->h, SEEK_CUR);
		ch++;
	}

		/* Alloc SDL_Surface and draw individual chars there */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	font->surface = SDL_CreateRGBSurface(SDL_SWSURFACE | SDL_SRCCOLORKEY, 
			totalwidth, maxheight, 32,
			0xff000000, 0x00ff0000, 0x0000ff00, 0x00000000);
#else
	font->surface = SDL_CreateRGBSurface(SDL_SWSURFACE | SDL_SRCCOLORKEY, 
			totalwidth, maxheight, 32,
			0x000000ff, 0x0000ff00, 0x00ff0000, 0x00000000);
#endif
	if(font->surface == NULL) {
		debug(sev_error, "bmf", "CreateRGBSurface failed:%s", SDL_GetError());
		goto err;
	}
	
	lseek(f, charsoffset, SEEK_SET);
	totalwidth = 0;
	ch = font->chars;
	for (i = 0; i < font->numchars; i++) {
		int x, y, idx;
		unsigned char *col;
		lseek(f, 6, SEEK_CUR);
		READ(buf, ch->w * ch->h);
		for (x = 0; x < ch->w; x++)
			for (y = 0; y < ch->h; y++) {
				idx = 3 * buf[ch->w*y + x];
				col = palette + idx;
				pixelRGBA(font->surface, totalwidth + x, y,
						col[0], col[1], col[2], 0xff);
			}
		totalwidth += ch->w;
		ch++;
	}
	SDL_SetColorKey(font->surface, SDL_SRCCOLORKEY,
			SDL_MapRGBA(font->surface->format,
				palette[0], palette[1], palette[2], 0x00)
			);
	tmp_surface = font->surface;
	font->surface = SDL_DisplayFormat(tmp_surface);
	SDL_FreeSurface(tmp_surface);
	return font;

err:
	if (font) {
		if (font->chars)
			free(font->chars);
		if (font->info)
			free(font->info);
		if (font->surface)
			SDL_FreeSurface(font->surface);
		free(font);
	}
	return NULL;
}

void
bmf_free(struct bmf_font *f)
{
	assert(f);
	SDL_FreeSurface(f->surface);
	free(f->chars);
	free(f->info);
	free(f);
}

struct bmf_char *
bmf_putc(SDL_Surface *s, struct bmf_font *f, int x, int y, char c)
{
	struct bmf_char *ch;
	SDL_Rect sr, dr;
	
	if (f->charmap[(int)c] == -1) {
		debug(sev_warn, "bmf", "character %c not in font.", c);
		return NULL;
	} else
		ch = f->chars + f->charmap[(int)c];
	sr.x = ch->offset; sr.y = 0; 
	sr.w = ch->w; sr.h = ch->h;
	dr.x = x + ch->relx; dr.y = y + ch->rely;
	dr.w = ch->w; dr.h = ch->h;
	SDL_BlitSurface(f->surface, &sr, s, &dr);
	return ch;
}

void
bmf_puts(SDL_Surface *s, struct bmf_font *f, int x, int y, const char *txt)
{
	struct bmf_char *ch;
	while (*txt) {
		ch = bmf_putc(s, f, x, y, *txt++);
		if (!ch)
			continue;
		x += ch->shift + f->add_space;
	}
}
